extern crate nihav_core;
extern crate nihav_allstuff;

use std::io::SeekFrom;
use std::fs::File;
use std::path::Path;
use nihav_core::io::byteio::{FileReader, ByteReader};
use nihav_core::frame::*;
use nihav_core::codecs::*;
use nihav_core::demuxers::*;
use nihav_core::detect;
use nihav_allstuff::*;
use std::env;

mod frmwriter;
use crate::frmwriter::*;

mod wavwriter;
use crate::wavwriter::WavWriter;

#[derive(Clone,Copy,PartialEq)]
enum NumberMode {
    Counter,
    PktPTS,
    FrmPTS,
}

struct FrameOutput {
    prefix:     &'static str,
    streamno:   usize,
    frameno:    u64,
    nmode:      NumberMode,
}

impl FrameOutput {
    fn output_frame(&mut self, pkt: &NAPacket, frm: NAFrameRef) {
        if frm.get_frame_type() != FrameType::Skip {
            let pts = match self.nmode {
                    NumberMode::Counter => { self.frameno },
                    NumberMode::PktPTS  => { pkt.get_pts().unwrap() },
                    NumberMode::FrmPTS  => { if let Some(pt) = frm.get_pts() { pt } else { pkt.get_pts().unwrap() } },
                };
            let vinfo = frm.get_buffer().get_video_info().unwrap();
            if vinfo.get_format().is_paletted() {
                write_palppm(self.prefix, self.streamno, pts, frm);
            } else if vinfo.get_format().get_model().is_yuv() {
                write_pgmyuv(self.prefix, self.streamno, pts, frm);
            } else if vinfo.get_format().get_model().is_rgb() {
                write_ppm(self.prefix, self.streamno, pts, frm);
            } else {
panic!(" unknown format");
            }
        }
        self.frameno += 1;
    }
}

struct AudioOutput {
    wwr:    WavWriter<'static>,
    wrote_header: bool,
}

impl AudioOutput {
    fn new(name: &String) -> Self { Self { wwr: WavWriter::new(name), wrote_header: false } }
    fn output_frame(&mut self, _pkt: &NAPacket, frm: NAFrameRef) {
        if !self.wrote_header {
            self.wwr.write_header(frm.get_info().as_ref().get_properties().get_audio_info().unwrap()).unwrap();
            self.wrote_header = true;
        }
        self.wwr.write_frame(frm.get_buffer()).unwrap();
    }
}

enum Outputter {
    Video(FrameOutput),
    Audio(AudioOutput),
    None,
}

fn main() {
    let args: Vec<_> = env::args().collect();

    if args.len() == 1 {
        println!("usage: nihav-tool [-noout] [-vn] [-an] input [lastpts]");
        return;
    }
    let mut lastpts: Option<u64> = None;
    let mut cur_arg: usize = 1;
    let mut noout = false;
    let mut decode_video = true;
    let mut decode_audio = true;
    let mut nmode = NumberMode::FrmPTS;

    while (cur_arg < args.len()) && (args[cur_arg].chars().next().unwrap() == '-') {
        match args[cur_arg].as_str() {
            "--"        => { break; },
            "-noout"    => { noout = true; },
            "-an"       => { decode_audio = false; },
            "-vn"       => { decode_video = false; },
            "-nm=count"  => { nmode = NumberMode::Counter; },
            "-nm=pktpts" => { nmode = NumberMode::PktPTS; },
            "-nm=frmpts" => { nmode = NumberMode::FrmPTS; },
            _           => { println!("unknown option {}", args[cur_arg]); return; },
        }
        cur_arg += 1;
    }
    let name = args[cur_arg].as_str();
    cur_arg += 1;
    if cur_arg < args.len() {
        lastpts = Some(u64::from_str_radix(args[cur_arg].as_str(), 10).unwrap());
    }

    let path = Path::new(name);
    let mut file = File::open(path).unwrap();
    let dmx_fact;
    let mut fr = FileReader::new_read(&mut file);
    let mut br = ByteReader::new(&mut fr);
    let res = detect::detect_format(name, &mut br);
    if res.is_none() {
        println!("cannot detect format for {}", name);
        return;
    }
    let (dmx_name, _) = res.unwrap();
println!("trying demuxer {} on {}", dmx_name, name);

    let mut dmx_reg = RegisteredDemuxers::new();
    nihav_register_all_demuxers(&mut dmx_reg);
    let mut dec_reg = RegisteredDecoders::new();
    nihav_register_all_codecs(&mut dec_reg);

    dmx_fact = dmx_reg.find_demuxer(dmx_name).unwrap();
    br.seek(SeekFrom::Start(0)).unwrap();
    let mut dmx = create_demuxer(dmx_fact, &mut br).unwrap();

    let mut decs: Vec<Option<Box<NADecoder>>> = Vec::new();
    let mut sids: Vec<u32> = Vec::new();
    let mut writers: Vec<Outputter> = Vec::new();
    for i in 0..dmx.get_num_streams() {
        let s = dmx.get_stream(i).unwrap();
        let info = s.get_info();
        let decfunc = dec_reg.find_decoder(info.get_name());
println!("stream {} - {} {}", i, s, info.get_name());
        let str_id = s.get_id();
        let mut has_out = false;
        sids.push(str_id);
        if info.is_video() {
            if decode_video {
                if decfunc.is_none() {
                    println!("no video decoder found!");
                    return;
                }
                let mut dec = (decfunc.unwrap())();
                dec.init(info).unwrap();
                decs.push(Some(dec));
                if !noout {
                    writers.push(Outputter::Video(FrameOutput{prefix: "", streamno: i, frameno: 1, nmode: nmode}));
                    has_out = true;
                }
            } else {
                decs.push(None);
            }
        } else if info.is_audio() {
            if decode_audio {
                if decfunc.is_none() {
                    println!("no audio decoder found!");
                    return;
                }
                let mut dec = (decfunc.unwrap())();
                dec.init(info).unwrap();
                decs.push(Some(dec));
                if !noout {
                    let name = format!("out{:02}.wav", i);
                    writers.push(Outputter::Audio(AudioOutput::new(&name)));
                    has_out = true;
                }
            } else {
                decs.push(None);
            }
        } else {
            decs.push(None);
panic!("decoder {} not found", info.get_name());
        }
        if !has_out {
            writers.push(Outputter::None);
        }
    }

    loop {
        let pktres = dmx.get_frame();
        if let Err(e) = pktres {
            if e == DemuxerError::EOF { break; }
        }
        let pkt = pktres.unwrap();
        let streamno = pkt.get_stream().get_id();
        let sr = sids.iter().position(|x| *x == streamno);
        let idx = sr.unwrap();
        if let Some(ref mut dec) = decs[idx] {
            let frm = dec.decode(&pkt).unwrap();
            if !noout {
                match writers[idx] {
                    Outputter::Video(ref mut wr) => { wr.output_frame(&pkt, frm); },
                    Outputter::Audio(ref mut wr) => { wr.output_frame(&pkt, frm); },
                    _ => {},
                };
            }
        }
        if pkt.get_pts() != None && lastpts.is_some() && pkt.get_pts() >= lastpts { break; }
    }
//panic!("end");
}
